/*
 *                        sc68 - xmms input plugin
 *         Copyright (C) 2001-2003 Benjamin Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/* Standard includes */
#include <glib.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* xmms includes */
#include "plugin.h"
#include "util.h"

#include <xmms68.h>

/* Generated config include.
 * Notice this is sc68 package file not xmms68 one !!
 */
#include <config68.h>

/* sc68 includes */
#include <api68/api68.h>
#include <file68/string68.h>

static api68_t * sc68 = 0;
static api68_init_t init68;

static InputPlugin xmms68_ip;
static volatile gboolean going;
static volatile gboolean audio_is_open;
static pthread_t play_thread;
static volatile int seek_to;

/* $$$ */
#ifndef _
# define _(A) (A)
#endif

/** Display to output debug statcked error messages.
 */
static void spool_error_message(void)
{
  const char * s;

  if (s = api68_error(), s) {
    api68_debug("Stacked Error Message:\n");
    do {
      api68_debug("%s\n", s);
    } while (s = api68_error(), s != NULL);
  }
}

/** Display a simple about dialog. */
static void xmms68_about(void)
{
  static GtkWidget *box;
  static char * text = 
    PACKAGE68_STRING " xmms plugin\n"
    "\n"
#ifdef XMMS68_DEBUG
    " !!! DEBUG VERSION !!! \n"
    "\n"
#endif       
    "Atari STF/E and Amiga sound emulator\n"
    "\n"
    "(c)1998-2003 Benjamin Gerard <ben@sashipa.com>\n"
    "\n";
  
  box = xmms_show_message(
    _("About xmms68"),
    _(text),
    _("Ok"),
    FALSE, NULL, NULL);
  gtk_signal_connect(GTK_OBJECT(box), "destroy", gtk_widget_destroyed, &box);
}


/** Set XMMS info. */
static void xmms68_set_info(int track)
{
  api68_music_info_t disk;
  api68_music_info_t mus;
  int no_disk, no_mus;
  gchar *name = 0 , *scroll = 0;

  no_disk = api68_music_info(sc68,&disk,0,0);
  if (no_disk) {
    return;
  }

  no_mus = api68_music_info(sc68,&mus,track,0) || !mus.track;
  if (no_mus) {
    name = g_strdup_printf("%s [%02u] / %s",
			   disk.title,
			   disk.tracks,
			   disk.author);
    scroll = g_strdup_printf("disk:%s [%s]"
			     "author:%s  composer:%s  "
			     "hardware:%s  replay routine:%s   ",
			     disk.title, disk.time,
			     disk.author, disk.composer,
			     disk.hwname,
			     disk.replay);
  } else {
    if (!SC68strcmp(disk.title,mus.title)) {
      name = g_strdup_printf("[%02u] %s / %s",
			     mus.track,
			     mus.title,
			     mus.author);
    } else {
      name = g_strdup_printf("%s [%02u] %s / %s",
			     disk.title,
			     mus.track,
			     mus.title,
			     mus.author);
    }
    scroll = g_strdup_printf("disk:%s [%s]  track:%s [%s]  "
			     "author:%s  composer:%s  "
			     "hardware:%s  replay routine:%s (%dHz)  ",
			     disk.title, disk.time,
			     mus.title, mus.time,
			     mus.author, mus.composer,
			     mus.hwname,
			     mus.replay, mus.rate);
  }

  xmms68_ip.set_info(name, disk.time_ms, 0, init68.sampling_rate, 2);
  if (name) {
    g_free(name);
  }

  if (scroll) {
    xmms68_ip.set_info_text(scroll);
    g_free(scroll);
  }
}


/** Test file format XMMS callback. */
static int xmms68_is_our_file(char *filename)
{
  int ret;
  api68_debug("xmms68_is_our_file(%s)\n", filename);
  ret = (api68_verify_file(filename) < 0) ? FALSE : TRUE;
  api68_debug(" -> %s\n", ret==TRUE?"TRUE":"FALSE");
  return ret;
}

static int xmms68_open_audio(void)
{
  int err = 0;
  if (audio_is_open) {
    api68_debug("xmms68_open_audio() : already open\n");
  } else if (!xmms68_ip.output ||
	     !xmms68_ip.output->open_audio(FMT_S16_LE,
					   init68.sampling_rate,
					   2)) {
    api68_debug("xmms68_open_audio() : ERROR\n");
    err = -1;
  } else { 
    audio_is_open = 1;
    api68_debug("xmms68_open_audio() : %dHz success\n", init68.sampling_rate);
  }
  return err;
}
    
static int xmms68_close_audio(void)
{
  int err = 0;
  if (!audio_is_open) {
    api68_debug("xmms68_close_audio() : already close\n");
  } else if (!xmms68_ip.output) {
    api68_debug("xmms68_close_audio() : no output !!\n");
    audio_is_open = 0;
  } else {
    api68_debug("xmms68_close_audio() : closing\n");
    xmms68_ip.output->close_audio();
    audio_is_open = 0;
  }
  return err;
}

/** XMMS stop callback.
 */
static void xmms68_stop(void)
{
  pthread_t p;

  api68_debug("xmms68_stop():\n -> going=%d thd=%p\n", going, play_thread);
  api68_debug("xmms68_stop():\n -> join %p\n", play_thread);
  api68_stop(sc68);
  going = 0;
  if (p = play_thread, p) {
    pthread_join(p, NULL);
    play_thread = 0;
    api68_debug("xmms68_stop():\n -> thd %p joined\n", p);
  }
  xmms68_close_audio();
}

static void *play_loop(void *arg)
{
  static char buffer[4*512];
  const int bytes   = sizeof(buffer);
  const int samples = sizeof(buffer) >> 2;
  int seek_pos;
  int code;

  code = api68_process(sc68, 0, 0);
  seek_pos = api68_seek(sc68,-1);
  if (seek_pos > 0 && xmms68_ip.output) {
    xmms68_ip.output->flush (seek_pos);
  }
  xmms68_set_info(-1);

  while (1) {

    if (seek_pos = seek_to, seek_pos != -1) {
      seek_to = -1;
      seek_pos = api68_seek(sc68, seek_pos);
      if (seek_pos != -1) {
	api68_debug("play_loop : flush ouput to %d\n", seek_pos);
	xmms68_ip.output->flush (seek_pos);
      }
    }

    code = api68_process(sc68, buffer, samples);
    if (code & API68_END) {
      break;
    }
    if (code & API68_CHANGE) {
      api68_debug("CHANGE\n");
      xmms68_set_info(-1);
    }
    xmms68_ip.add_vis_pcm(xmms68_ip.output->written_time(),
			  FMT_S16_LE, 2, bytes, buffer);
    while (xmms68_ip.output->buffer_free() < bytes && going) {
      xmms_usleep(30000);
    }
    if (going) {
      xmms68_ip.output->write_audio(buffer, bytes);
    }
    if (going == 2) {
      going = 1;
    }
  }
  api68_debug("xmms68: out of play loop (code=%08x)\n", code);
  going = 0;
  xmms68_set_info(0);
  api68_close(sc68);

  spool_error_message();

  /* Make sure the output plugin stops prebuffering. */
  if (audio_is_open) {
    xmms68_ip.output->buffer_free();
    xmms68_ip.output->buffer_free();
  }
  pthread_exit(NULL);
}


static void xmms68_play(char *filename)
{
  api68_debug("xmms68_play(%s)\n", filename);

  if (play_thread) {
    api68_debug("xmms68_play(%s):\n -> play_thread %p exist !!!\n",
		filename, play_thread);
    return;
  }

  if (going) {
    api68_debug("xmms68_play(%s):\n -> already going !!\n", filename);
    return;
  }

  if (api68_load_file(sc68, filename) < 0) {
    goto error;
  }

  if (xmms68_open_audio() < 0) {
    goto error;
  }

  going = 2;
  seek_to = -1;
  pthread_create(&play_thread, NULL, play_loop, 0);
  api68_debug("xmms68_play(%s):\n -> thd=%p\n",filename,play_thread);
  while (going == 2) {
    api68_debug("xmms68_play(%s):\n -> wait really going (%d)\n",
		  filename, going);
    xmms_usleep(30000);
  }

  return;

error:
  going = FALSE;
  spool_error_message();
}

/** XMMS pause callback.
 *
 *    This function just call the output plugin pause function.
 */
static void xmms68_pause(short paused)
{
  if (audio_is_open && xmms68_ip.output && xmms68_ip.output->pause) {
    xmms68_ip.output->pause(paused);
  }
}

static int xmms68_get_time(void)
{
  int t;

  if (!audio_is_open) {
    return -2;
  }

  if (!going && !xmms68_ip.output->buffer_playing()) {
    t = -1;
  } else {
    t = xmms68_ip.output->output_time();
  }

/*   api68_debug("xmms68_get_time() -> %d\n",t); */
  return t;
}

static void xmms68_song_info(char *filename, char **title, int *length)
{
  api68_disk_t d;
  api68_music_info_t disk;

  api68_debug("xmms68_song_info(%s)\n", filename);

  *length = -1;
  *title = 0;

  d = api68_load_disk_file(filename);
  if (d && !api68_music_info(sc68, &disk,0,d)) {
    gchar *name;

    api68_debug("xmms68_song_info(%s) : set disk info\n");
    name =
        g_strdup_printf("%s [%02u] / %s",
			disk.title, disk.tracks, disk.author);
    *title = name;
    *length = disk.time_ms;

  } else {

    char *s1, *s2;

    /* If if don't do that xmms call is_our_file() and this function
       all the time which can be very annoying for gzipped file.
    */
    api68_debug("xmms68_song_info(%s) : set default\n");

    s1 = strrchr(filename,'/');
    s2 = strrchr(filename,'\\');
    if (s2 > s1) {
      s1 = s2;
    }
    s1 = s1 ? s1+1 : filename;
    s1 = g_strdup(s1);
    
    if (s1) {
      s2 = strrchr(s1,'.');
      if (s2) {
	  *s2 = 0;
	  if (!SC68strcmp(s2+1,"gz")) {
	    s2 = strrchr(s1,'.');
	    if (s2) {
	      *s2 = 0;
	    }
	  }
      }
      *title = s1;
    }
  }
  api68_free(d);
}

/** XMMS plugin description. */
static char xmms68_description[] =
"Atari-ST and Amiga soundchip emulator " VERSION68;

/* Called when the plugin is loaded
*/
static void xmms68_init(void)
{
  api68_debug("\nxmms68_init() : enter\n");
  play_thread = 0;
  going = 0;

  memset(&init68, 0, sizeof(init68));
  init68.alloc = malloc;
  init68.free = free;
#if defined(XMMS68_DEBUG ) || defined(_DEBUG)
  init68.debug = vfprintf;
  init68.debug_cookie = stderr;
#endif

  sc68 = api68_init(&init68);
}

/** Clean exit XMMS callback. */
static void xmms68_cleanup(void)
{
  api68_debug("\nxmms68_cleanup: enter\n");
  api68_shutdown(sc68);
  api68_debug("\nxmms68_cleanup: Success\n");
}

/** Seek XMMS callback.
 *
 *    This is used to change track in the current disk.
 */
static void xmms68_seek(int time)
{
  if (going) {
    seek_to = time*1000;
  }
}

/** XMMS input plugin instance. */
static InputPlugin xmms68_ip =
{
  NULL,                 /* !! Handle */
  NULL,                 /* !! Filename */
  xmms68_description,   /* Description */
  xmms68_init,          /* Init */
  xmms68_about,         /* Show about */
  NULL,                 /* Configure */
  xmms68_is_our_file,
  NULL,                 /* Scan dir */
  xmms68_play,
  xmms68_stop,
  xmms68_pause,
  xmms68_seek,          /* Seek */
  NULL,                 /* Set eq */
  xmms68_get_time,      /* */
  NULL,                 /* get volume l/r */
  NULL,                 /* set volume l/r */
  xmms68_cleanup,       /* Exit call */
  NULL,                 /* Get vis type (OBSOLETE) */
  NULL,                 /* Sends data to vis pluggin */
  NULL,                 /* !!! Set info */
  NULL,                 /* !!! Set song title */
  xmms68_song_info,
  NULL,                 /* file info box */
  NULL                  /* Output plugin */
};

/** Get input plugin info XMMS callback. */
InputPlugin *get_iplugin_info(void)
{
  return &xmms68_ip;
}
